#!/bin/bash

# Copyright (c) 2015-2017, NVIDIA CORPORATION.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#  * Neither the name of NVIDIA CORPORATION nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#
# tegrafuse.sh: Read/Write fuse in the target board.
#    tegrafuse.sh runs in the target board.
#
# Usage: Place the board in recovery mode and run:
#
#    sudo ./tegrafuse.sh [options] [fusename]
#    sudo ./tegrafuse.sh [options] <fusename> <fusevalue>
#
#    more detail enter './tegrafuse.sh -h'
#
# Examples:
#	sudo ./tegrafuse.sh 	           --> Display all Tegra fuses.
#	sudo ./tegrafuse odm_reserved 0x2  --> program odm_reserver to 0x2.
#

declare -a fusetab=( \
	"device_key" \
	"jtag_disable" \
	"odm_lock" \
	"odm_production_mode" \
	"odm_reserved" \
	"pkc_disable" \
	"public_key" \
	"sec_boot_dev_cfg" \
	"secure_boot_key" \
	"sw_reserved" \
);

rd_fuse ()
{
	local i;
	local nm=$1;
	for (( i=0; i<${#fusetab[@]}; i++ )); do
		if [ "${nm}" != "" -a "${nm}" != "${fusetab[$i]}" ]; then
			continue;
		fi;
		if [ ! -f ${sdir}/${fusetab[$i]} ]; then
			echo "Unsupported fuse: ${fusetab[$i]}";
			continue;
		fi;
		local val=`cat ${sdir}/${fusetab[$i]}`;
		if [ "${nm}" != "" ]; then
			echo "${val}";
			return;
		fi;
		echo "${fusetab[$i]} : ${val}";
	done;
	if [ "${nm}" != "" ]; then
		echo "Unknown fuse name: ${nm}";
		exit 1;
	fi;
}

wr_fuse ()
{
	local i;
	local nm=$1;
	local val=$2;

	if [ "${nm}" = "" ]; then
		echo "Missing fuse name.";
		exit 1;
	fi;
	if [ "${val}" = "" ]; then
		echo "Missing fuse value for ${nm}.";
		exit 1;
	fi;

	for (( i=0; i<${#fusetab[@]}; i++ )); do
		if [ "${nm}" != "${fusetab[$i]}" ]; then
			continue;
		fi;
		if [ ! -f ${sdir}/${fusetab[$i]} ]; then
			echo "Unsupported fuse: ${fusetab[$i]}";
			exit 1;
		fi;
		if [ ! -w ${sdir}/${fusetab[$i]} ]; then
			echo "Permission denied: ${fusetab[$i]}";
			exit 1;
		fi;
		echo "${val}" > ${sdir}/${fusetab[$i]};
		return;
	done;
	echo "Unknown fuse name: ${nm}";
	exit 1;
}

usage ()
{
	cat << EOF
usage: sudo ${progname} [options] [fusename]
       sudo ${progname} [options] <fusename> <fusevalue>
    options:
        -d -- sysfs directory name. (default=/sys/devices/platform/tegra-fuse)
        -h -- help
EOF
	exit 1;
}

sdir="/sys/devices/platform/tegra-fuse";	# T210 dir
progname=$0;
me=`whoami`;
if [ "${me}" != "root" ]; then
	echo "${progname} requires root privilege.";
	exit 1;
fi;

while getopts "d:h" OPTION
do
	case $OPTION in
	d) sdir=${OPTARG}; ;;
	*) usage; ;;
	esac
done
shift $((OPTIND - 1));
if [ $# -gt 1 ]; then
	wflag="1";
	fuseval=$2;
fi;
if [ $# -gt 0 ]; then
	fusename=$1;
fi;

if [ ! -d "${sdir}" ]; then
	sdir="/sys/devices/3820000.efuse";
fi;

if [ "${wflag}" != "" ]; then
	wr_fuse ${fusename} ${fuseval};
else
	rd_fuse ${fusename};
fi;
exit 0;
